"use strict";

// Box that can contain other components
Toolkit.Panel = class Panel extends Toolkit.Component {

    // Object constructor
    constructor(application, options) {
        super(application, "div", options);

        // Configure instance fields
        options = options || {};
        this.alignCross  = "start";
        this.alignMain   = "start";
        this.application = application;
        this.children    = [];
        this.columns     = null;
        this.direction   = "row";
        this.layout      = null;
        this.overflowX   = options.overflowX || "hidden";
        this.overflowY   = options.overflowY || "hidden";
        this.rows        = null;
        this.wrap        = false;

        // Configure element
        this.element.style.minHeight = "0";
        this.element.style.minWidth  = "0";
        this.setOverflow(this.overflowX, this.overflowY);

        // Configure layout
        options = options || {};
        this.setLayout(options.layout || null, options);
    }



    ///////////////////////////// Public Methods //////////////////////////////

    // Add a component as a child of this container
    add(component, index) {

        // Determine the ordinal position of the element within the container
        index = !(typeof index == "number") ? this.children.length :
            Math.floor(Math.min(Math.max(0, index), this.children.length));

        // Add the component to the container
        let ref = this.children[index] || null;
        component.parent = this;
        this.element.insertBefore(component.element,
            ref == null ? null : ref.element);
        this.children.splice(index, 0, component);

        return component;
    }

    // Create a Button and associate it with the application
    newButton(options) {
        return new Toolkit.Button(this.application, options);
    }

    // Create a MenuBar and associate it with the application
    newMenuBar(options) {
        return new Toolkit.MenuBar(this.application, options);
    }

    // Create a Panel and associate it with the application
    newPanel(options) {
        return new Toolkit.Panel(this.application, options);
    }

    // Create a Window and associate it with the application
    newWindow(options) {
        return new Toolkit.Window(this.application, options);
    }

    // Determine the index of the next visible child
    nextChild(index) {
        for (let x = 0; x <= this.children.length; x++) {
            index = (index + 1) % this.children.length;
            let comp = this.children[index];
            if (comp.isVisible())
                return index;
        }
        return -1;
    }

    // Determine the index of the previous visible child
    previousChild(index) {
        for (let x = 0; x <= this.children.length; x++) {
            index = (index + this.children.length - 1) % this.children.length;
            let comp = this.children[index];
            if (comp.isVisible())
                return index;
        }
        return -1;
    }

    // Remove a component from the container
    remove(component) {

        // Locate the component in the children
        let index = this.children.indexOf(component);
        if (index == -1)
            return;

        // Remove the component
        component.parent = null;
        component.element.remove();
        this.children.splice(index, 1);
    }

    // Configure the element's layout
    setLayout(layout, options) {

        // Configure instance fields
        this.layout = layout;

        // Processing by layout
        switch (layout) {
            case "block"  : this.setBlockLayout  (options); break;
            case "desktop": this.setDesktopLayout(options); break;
            case "flex"   : this.setFlexLayout   (options); break;
            case "grid"   : this.setGridLayout   (options); break;
            default       : this.setNullLayout   (options); break;
        }

    }

    // Configure the panel's overflow scrolling behavior
    setOverflow(x, y) {
        this.overflowX = x || "hidden";
        this.overflowY = y || "hidden";
        this.element.style.overflow = this.overflowX + " " + this.overflowY;
    }

    // Specify the semantic role of the panel
    setRole(role) {
        if (!role)
            this.element.removeAttribute("role");
        else this.element.setAttribute("role", "" + role);
    }



    ///////////////////////////// Private Methods /////////////////////////////

    // Resize event handler
    onresize(desktop) {

        // Error checking
        if (this.layout != "desktop")
            return;

        // Ensure all child windows are visible in the viewport
        for (let wnd of this.children) {
            let bounds = wnd.getBounds();
            wnd.contain(
                desktop,
                bounds,
                bounds.x - desktop.x,
                bounds.y - desktop.y
            );
        }

    }

    // Configure a block layout
    setBlockLayout(options) {

        // Configure instance fields
        this.layout = "block";

        // Configure element
        this.setDisplay("block");
    }

    // Configure a desktop layout
    setDesktopLayout(options) {

        // Configure instance fields
        this.layout = "desktop";

        // Configure element
        this.setDisplay("block");
        this.element.style.position = "relative";
        if (this.resizeObserver == null)
            this.addResizeListener(b=>this.onresize(b));
    }

    // Configure a flex layout
    setFlexLayout(options) {

        // Configure instance fields
        this.alignCross = options.alignCross || "start";
        this.alignMain  = options.alignMain  || "start";
        this.direction  = options.direction  || this.direction;
        this.layout     = "flex";
        this.wrap       = !!options.wrap;

        // Working variables
        let alignCross = this.alignCross;
        let alignMain  = this.alignMain;
        if (alignCross == "start" || alignCross == "end")
            alignCross = "flex-" + alignCross;
        if (alignMain == "start" || alignMain == "end")
            alignMain = "flex-" + alignMain;

        // Configure element
        this.setDisplay("flex");
        this.element.style.alignItems     = alignCross;
        this.element.style.flexDirection  = this.direction;
        this.element.style.justifyContent = alignMain;
        this.element.style.flexWrap = this.wrap ? "wrap" : "nowrap";
    }

    // Configure a grid layout
    setGridLayout(options) {

        // Configure instance fields
        this.columns = options.columns   || null;
        this.layout  = "grid";
        this.rows    = options.rows      || null;

        // Configure element
        this.setDisplay("grid");
        if (this.columns == null)
            this.element.style.removeProperty("grid-template-columns");
        else this.element.style.gridTemplateColumns = this.columns;
        if (this.rows == null)
            this.element.style.removeProperty("grid-template-rows");
        else this.element.style.gridTemplateRows = this.rows;
    }

    // Configure a null layout
    setNullLayout(options) {

        // Configure instance fields
        this.layout = null;

        // Configure element
        this.setDisplay(null);
        this.element.style.removeProperty("align-items"          );
        this.element.style.removeProperty("flex-wrap"            );
        this.element.style.removeProperty("grid-template-columns");
        this.element.style.removeProperty("grid-template-rows"   );
        this.element.style.removeProperty("justify-content"      );
        this.element.style.removeProperty("flex-direction"       );
        this.element.style.removeProperty("overflow-x"           );
        this.element.style.removeProperty("overflow-y"           );
    }

};
